<?php
/* vim: set expandtab tabstop=4 shiftwidth=4 softtabstop=4: */
// This is a IA (chatgpt) powered conversational Bot ussing the Assistant API.
// You must supply your OpenAI API key and an assistant id for it to work. 
// You must create an assistant into OpenAI however you like. To get the 
// Assistant ID you can test your assistant in the Playground and you will 
// see the id (asst_xxxxxx) below the name.
//
// There are two special comands that the AI can use to handle conversations:
// [transfer] and [close]. You can add the following sentence to the assistnat
// system prompt to handle those cases:
//
// "If the user is angry, frustated or whats to talk to a real person, respond 
// saying that you will pass the conversation to a human and add the command 
// [transfer] in your reply, when finishing the conversation add the command
// [close]."
//
// There is also a GUARDRAILS_PROMPT that you can use to append anything to 
// the user question/prompt. It can be used to restrict answers further.

$DEBUG        = 0;
$OPENAI_TOKEN = '';
$ASSISTANT_ID = '';
$GUARDRAILS_PROMPT = '';

// override all hardcoded options if we have a bot name and ini file
if(isset($_REQUEST['bot'])) {
    $filename = getcwd()."/".$_REQUEST['bot'].".ini";
    if(is_readable($filename)) {
        $ini = parse_ini_file($filename,false);
        $OPENAI_TOKEN = $ini['openai_token'];
        $ASSISTANT_ID = $ini['assistant_id'];
        $GUARDRAILS_PROMPT = $ini['guardrails_prompt'];
        $FUNCTIONS = $ini['functions'];
    }
}

if($OPENAI_TOKEN=='' || $ASSISTANT_ID=='') {
    $response = "Bot is not yet configured, be sure to set up your OpenAI API key";
    send_response($response); // dies after sending response
}

$tools = array();
$dorequest=1;
$filename_tools = getcwd()."/".$_REQUEST['bot']."_tools.json";
if(is_readable($filename_tools)) {
    $last = filemtime($filename_tools);
    $now = time();
    $dif = $now-$last;
    if($dif>360) {
        $dorequest=1;
    } else {
        $dorequest=0;
    }
} else {
    $dorequest=1;
}
if($dorequest==1) {
    $tools = openai_get_assistant_tools($OPENAI_TOKEN,$ASSISTANT_ID);
    file_put_contents($filename_tools,json_encode($tools));
} else {
    $tools = json_decode(file_get_contents($filename_tools),true);
}

log_debug(print_r($tools,1));

$json = file_get_contents('php://input');
$data = json_decode($json,1);

log_debug("POST RECEIVED:\n".print_r($data,1));

$current_state = '';

// For assistants, the bot state is the openai thread
if($data['state']=='') {
    // No state, its a new conversation, create a thread
    $thread = openai_create_thread($OPENAI_TOKEN,$tools);
    $current_state = array("thread"=>$thread);
} else {
    $bot_state = json_decode($data['state'],1);
    $thread = $bot_state['thread'];
}
 
if($thread=='' || $thread === null) {
    $response = "Digital brain is not working correctly (A01) [transfer]";
    send_response($response); // dies after sending response
}

$user_prompt = $data['body'];
$user_prompt.=". ".$GUARDRAILS_PROMPT; // add guardrails to question

$msgid = openai_create_message($OPENAI_TOKEN,$thread,$user_prompt);

if($msgid=='' || $msgid === null) {
    $response = "Digital brain is not working correctly (A02) [transfer]";
    send_response($response); // dies after sending response
}

$runid = openai_create_run($OPENAI_TOKEN,$ASSISTANT_ID,$thread);
if($runid=='' || $runid === null) {
    $response = "Digital brain is not working correctly (A03) [transfer]";
    send_response($response); // dies after sending response
}

// Get run status up to 30 attempts/seconds
$ok=0;
$inc=1000000;
for($a=1;$a<60;$a++) {
    if ($a % 10 == 0) {
        $inc = (1000000 + ($a / 10) * 500000);
    }
    usleep($inc);
    $data = openai_retrieve_run($OPENAI_TOKEN,$thread,$runid);
    $status = $data['status'];
    log_debug("get run status $status");
    if($status=='completed') { 
        $ok=1; 
        break; 
    } else if($status=='requires_action') {
        $return = call_functions($OPENAI_TOKEN,$thread,$runid,$data['required_action']['submit_tool_outputs']['tool_calls']);
   } 
}

if($ok==0) {
    $response = "Digital brain is not working correctly (A04) [transfer]";
    send_response($response); // dies after sending response
}

// Retrieve first one from list (desc order limit 2)
$messages = openai_list_messages($OPENAI_TOKEN,$thread);
$openai_assistant_reply =  $messages[0]['content'][0]['text']['value'];

$command = extract_command_from_reply($openai_assistant_reply);
if($command=='close' || $command=='transfer') {
    openai_delete_thread($OPENAI_TOKEN,$thread);
    log_debug("OpenAI delete thread $thread");
}

$final_msg = remove_annotations($openai_assistant_reply);
send_response($final_msg);
// END OF PROGRAM

function openai_create_thread($openai_token,$tools) {
    $url = 'https://api.openai.com/v1/threads';
    $final_tools = array("tool_resources"=>$tools);
    $data_string = json_encode($final_tools);
    $headers=array('Content-Type: application/json','Content-Length: ' . strlen($data_string),"Authorization: Bearer $openai_token","OpenAI-Beta: assistants=v2");
    $ch = curl_init($url);
    curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
    curl_setopt($ch, CURLOPT_HEADER, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    curl_setopt($ch, CURLOPT_TIMEOUT, 5);
    curl_setopt($ch, CURLOPT_VERBOSE, true);
    curl_setopt($ch, CURLOPT_SSLVERSION, 1);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    curl_setopt($ch, CURLOPT_POST, 1);
    curl_setopt($ch, CURLOPT_POSTFIELDS, $data_string);
    $return = curl_exec($ch);
    curl_close($ch);
    $response = json_decode($return,1);
    log_debug("create thread response: ".print_r($response,1));
    return $response['id'];
}

function openai_delete_thread($openai_token,$threadid) {
    $url = 'https://api.openai.com/v1/threads/'.$threadid;
    $headers=array('Content-Type: application/json',"Authorization: Bearer $openai_token","OpenAI-Beta: assistants=v2");
    $ch = curl_init($url);
    curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
    curl_setopt($ch, CURLOPT_HEADER, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    curl_setopt($ch, CURLOPT_TIMEOUT, 5);
    curl_setopt($ch, CURLOPT_VERBOSE, true);
    curl_setopt($ch, CURLOPT_SSLVERSION, 1);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    curl_setopt($ch, CURLOPT_CUSTOMREQUEST, "DELETE");
    $return = curl_exec($ch);
    curl_close($ch);
    $response = json_decode($return,1);
    log_debug("delete trhead response: ".print_r($response,1));
    return $response['deleted'];
}

function openai_create_message($openai_token,$threadid,$content) {
    $url = 'https://api.openai.com/v1/threads/'.$threadid."/messages";
    $data = array(
       "role"=> "user",
       "content" => $content
    );
    $data_string = json_encode($data);
    $headers=array('Content-Type: application/json','Content-Length: ' . strlen($data_string),"Authorization: Bearer $openai_token","OpenAI-Beta: assistants=v2");
    $ch = curl_init($url);
    curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
    curl_setopt($ch, CURLOPT_HEADER, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    curl_setopt($ch, CURLOPT_TIMEOUT, 5);
    curl_setopt($ch, CURLOPT_VERBOSE, true);
    curl_setopt($ch, CURLOPT_SSLVERSION, 1);
    curl_setopt($ch, CURLOPT_POST, 1);
    curl_setopt($ch, CURLOPT_POSTFIELDS, $data_string);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    $return = curl_exec($ch);
    curl_close($ch);
    $response = json_decode($return,1);
    log_debug("create message url: ".$url);
    log_debug("create message payload: ". $data_string);
    log_debug("create message response: ".print_r($response,1));
    return $response['id'];
}

function openai_list_messages($openai_token,$threadid) {
    $url = 'https://api.openai.com/v1/threads/'.$threadid."/messages?order=desc&limit=2";
    $headers=array('Content-Type: application/json',"Authorization: Bearer $openai_token","OpenAI-Beta: assistants=v2");
    $ch = curl_init($url);
    curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
    curl_setopt($ch, CURLOPT_HEADER, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    curl_setopt($ch, CURLOPT_TIMEOUT, 5);
    curl_setopt($ch, CURLOPT_VERBOSE, true);
    curl_setopt($ch, CURLOPT_SSLVERSION, 1);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    $return = curl_exec($ch);
    curl_close($ch);
    $response = json_decode($return,1);
    log_debug("create message response: ".print_r($response,1));
    return $response['data'];
}

function openai_create_run($openai_token,$assistantid,$threadid) {
    $url = 'https://api.openai.com/v1/threads/'.$threadid."/runs";
    $data = array(
       "assistant_id"=> $assistantid
    );
    $data_string = json_encode($data);
    $headers=array('Content-Type: application/json','Content-Length: ' . strlen($data_string),"Authorization: Bearer $openai_token","OpenAI-Beta: assistants=v2");
    $ch = curl_init($url);
    curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
    curl_setopt($ch, CURLOPT_HEADER, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    curl_setopt($ch, CURLOPT_TIMEOUT, 5);
    curl_setopt($ch, CURLOPT_VERBOSE, true);
    curl_setopt($ch, CURLOPT_SSLVERSION, 1);
    curl_setopt($ch, CURLOPT_POSTFIELDS, $data_string);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    curl_setopt($ch, CURLOPT_POST, 1);
    $return = curl_exec($ch);
    curl_close($ch);
    $response = json_decode($return,1);
    log_debug("create run response: ".print_r($response['status'],1));
    return $response['id'];
}

function openai_retrieve_run($openai_token,$threadid,$runid) {
    $url = 'https://api.openai.com/v1/threads/'.$threadid."/runs/".$runid;
    $headers=array('Content-Type: application/json',"Authorization: Bearer $openai_token","OpenAI-Beta: assistants=v2");
    $ch = curl_init($url);
    curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
    curl_setopt($ch, CURLOPT_HEADER, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    curl_setopt($ch, CURLOPT_TIMEOUT, 5);
    curl_setopt($ch, CURLOPT_VERBOSE, true);
    curl_setopt($ch, CURLOPT_SSLVERSION, 1);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    $return = curl_exec($ch);
    curl_close($ch);
    $response = json_decode($return,1);
    log_debug("retrieve run response: ".print_r($response['status'],1));
    return $response;
}

function openai_get_assistant_tools($openai_token,$assistant_id) {
    $url = 'https://api.openai.com/v1/assistants/'.$assistant_id;
    $headers=array('Content-Type: application/json',"Authorization: Bearer $openai_token","OpenAI-Beta: assistants=v2");
    $ch = curl_init($url);
    curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
    curl_setopt($ch, CURLOPT_HEADER, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    curl_setopt($ch, CURLOPT_TIMEOUT, 5);
    curl_setopt($ch, CURLOPT_VERBOSE, true);
    curl_setopt($ch, CURLOPT_SSLVERSION, 1);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    $return = curl_exec($ch);
    curl_close($ch);
    $response = json_decode($return,1);
    log_debug("get assistant return: ".print_r($response,1));
    return $response['tool_resources'];
}

function openai_submit_tool_output($openai_token,$threadid,$runid,$content) {
    $url = 'https://api.openai.com/v1/threads/'.$threadid."/runs/".$runid."/submit_tool_outputs";
    $data = array(
       "tool_outputs"=> $content
    );
    $data_string = json_encode($data);
    $headers=array('Content-Type: application/json','Content-Length: ' . strlen($data_string),"Authorization: Bearer $openai_token","OpenAI-Beta: assistants=v2");
    $ch = curl_init($url);
    curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
    curl_setopt($ch, CURLOPT_HEADER, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    curl_setopt($ch, CURLOPT_TIMEOUT, 5);
    curl_setopt($ch, CURLOPT_VERBOSE, true);
    curl_setopt($ch, CURLOPT_SSLVERSION, 1);
    curl_setopt($ch, CURLOPT_POST, 1);
    curl_setopt($ch, CURLOPT_POSTFIELDS, $data_string);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
    $return = curl_exec($ch);
    curl_close($ch);
    $response = json_decode($return,1);
    log_debug("submit tool output url: ".$url);
    log_debug("submit tool output payload: ". $data_string);
    log_debug("submit tool output response: ".print_r($response,1));
    return $response['id'];
}

function call_functions($token,$thread,$runid,$tools) {
    global $FUNCTIONS;
    $func=array();
    $funcr = json_decode(base64_decode($FUNCTIONS),1);
    foreach($funcr as $idx=>$funcdata) {
       $func[$funcdata['name']]=$funcdata['url'];
    }

    $final_response = array();

    foreach($tools as $idx=>$data) {
        $tool_call_id = $data['id'];
        $func_name    = $data['function']['name']; 
        $func_args    = $data['function']['arguments'];
        if(isset($func[$func_name])) {
            log_debug("call function $func_name with args ".$func_args);
            $output = post_json_request($func[$func_name],$func_args);
            log_debug("response: $output");
        } else {
            log_debug("requested function $func_name is not defined in bot configuration");
            $output = array("success"=>"false");
        }
        $response = array("tool_call_id"=>$tool_call_id,"output"=>$output);
        $final_response[]=$response;
    }
    log_debug("about to submit output");
    log_debug(print_r($final_response,1));
    openai_submit_tool_output($token,$thread,$runid,$final_response);
}

function post_json_request($url,$payload) {
    log_debug("do post to $url");
    $ch = curl_init( $url );
    curl_setopt( $ch, CURLOPT_POSTFIELDS, $payload );
    curl_setopt( $ch, CURLOPT_HTTPHEADER, array('Content-Type:application/json'));
    curl_setopt( $ch, CURLOPT_RETURNTRANSFER, true );
    curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
    curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
    $result = curl_exec($ch);
    log_debug("result: $result");
    curl_close($ch);
    return $result;
}

function extract_command_from_reply($text) {
    $pattern = '/\[(.*?)\]/';
    preg_match($pattern, $text, $matches);
    $extractedString = (isset($matches[1])) ? $matches[1] : '';
    return $extractedString;
}

function remove_annotations($text) {
    $pattern = '/【(.*?)】/';
    $replacedString = preg_replace($pattern, '', $text);
    return $replacedString;
}

function send_response($response) {
    global $current_state;
    $return   = array("body"=>$response,"attach"=>'');
    if($current_state!='') {
        $return['state']=$current_state;
    }
    log_debug("REPLY: ".print_r($return,1));
    echo json_encode($return);
    die();
}

function log_debug($text) {
    global $DEBUG;
    $logfile = basename(__FILE__,'.php');
    if($DEBUG==1) {
        file_put_contents("/tmp/${logfile}.log",$text."\n",FILE_APPEND);
    }
}
